#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <cpgplot.h>
#include <omp.h>
#include <complex.h>
#include "pca_utils.h"

#include "mandel.h"

#define AUTOSCALE -123456

/*----------------------------------------------------------------------*/

void display_map(float **map, int npix, int fmin, int fmax, double ind)
{
  static int firsttime=1;
  if (firsttime) {
    cpgopen("/xserve");
    cpgenv(1,npix,1,npix,0,1);
    my_pgctab(0,1,0,1,1,0,256);
    firsttime=1;
  }

  if (fmin==AUTOSCALE){
    fmin=map[0][0];
    for (int i=0;i<npix;i++)
      for (int j=0;j<npix;j++)
	if (map[i][j]<fmin)
	  fmin=map[i][j];
  }

  if (fmax==AUTOSCALE){
    fmax=map[0][0];
    for (int i=0;i<npix;i++)
      for (int j=0;j<npix;j++)
	if (map[i][j]>fmax)
	  fmax=map[i][j];
  }

  float **map_copy=fmatrix(npix,npix);
  for (int i=0;i<npix;i++)
    for (int j=0;j<npix;j++)
      map_copy[i][j]=pow(map[j][i]-fmin,ind); //take transpose for plotting purposes
  fmax=pow(fmax-fmin,ind);
  //fmin=pow(fmin,ind);
  fmin=0;

  float tr[6];
  tr[0]=0;tr[1]=1;tr[2]=0;tr[3]=0;tr[4]=0;tr[5]=1;
  cpgimag(map_copy[0],npix,npix,1,npix,1,npix,fmin,fmax,tr);
  free(map_copy[0]);
  free(map_copy);
}

/*----------------------------------------------------------------------*/

char display_map_new_limits(float **map, int npix, int fmin, int fmax, double ind, double *xmin, double *xmax, double *ymin, double *ymax)
{ 
  static int firsttime=1;
  if (firsttime) {
    cpgopen("/xserve");
    cpgenv(1,npix,1,npix,1,-2);  
    my_pgctab(0,1,0,1,1,0,256);
    firsttime=0;
  }

  float tr[6] = {0.0, 1.0, 0.0, 0.0, 0.0, 1.0};
  if (fmin==AUTOSCALE){
    fmin=map[0][0];
    for (int i=0;i<npix;i++)
      for (int j=0;j<npix;j++)
	if (map[i][j]<fmin)
	  fmin=map[i][j];
  } 
  
  if (fmax==AUTOSCALE){
    fmax=map[0][0];
    for (int i=0;i<npix;i++)
      for (int j=0;j<npix;j++)
	if (map[i][j]>fmax)
	  fmax=map[i][j];
  }

  float **map_copy=fmatrix(npix,npix);
  for (int i=0;i<npix;i++)
    for (int j=0;j<npix;j++)
      map_copy[i][j]=pow(map[j][i],ind); //take transpose for plotting purposes
  fmin=pow(fmin,ind);
  fmax=pow(fmax,ind);

  cpgimag(map_copy[0],npix,npix,1,npix,1,npix,fmin,fmax,tr);
  free(map_copy[0]);
  free(map_copy);
  

  float fx1in, fy1in, fx2in, fy2in;
  char press;
  cpgcurs(&fx1in,&fy1in,&press);
  if (press=='X') {
    *xmin=(fx1in/npix)*(*xmax-*xmin)+(*xmin);
    *ymin=(fy1in/npix)*(*ymax-*ymin)+(*ymin);
    return press;
  }
  cpgsci(3);  
  cpgband(2,0,fx1in,fy1in,&fx2in,&fy2in,&press);

  //fprintf(stderr,"limits are %10.3f %10.3f, %10.3f %10.3f\n",fx1in,fx2in,fy1in,fy2in);


  double dfx1in=(fx1in/npix)*(*xmax-*xmin)+(*xmin);
  double dfx2in=(fx2in/npix)*(*xmax-*xmin)+(*xmin);
  double dfy1in=(fy1in/npix)*(*ymax-*ymin)+(*ymin);
  double dfy2in=(fy2in/npix)*(*ymax-*ymin)+(*ymin);

  double dx=fabs(dfx2in-dfx1in);
  double dy=fabs(dfy2in-dfy1in);
  if (dy>dx)
    dx=dy;
  double xcent=0.5*(dfx1in+dfx2in);
  double ycent=0.5*(dfy1in+dfy2in);

  *xmin=xcent-0.5*dx;
  *xmax=*xmin+dx;
  *ymin=ycent-0.5*dx;
  *ymax=*ymin+dx;
  
  //printf("current limits are %8.4f %8.4f %8.4f %8.4f\n",*xmin,*xmax,*ymin,*ymax);
  //if (press=='X')
  //rexit(EXIT_SUCCESS);
  return press;


}

/*----------------------------------------------------------------------*/

int how_many_iter(double complex a, int maxiter)
{
  int i=0;
  double complex b=a;
  double r;
  while (i<maxiter) {
    r=cimag(b)*cimag(b)+creal(b)*creal(b);
    if (r>4)
      return i;
    b=b*b+a;


    i++;
  }
  return maxiter;
}

/*----------------------------------------------------------------------*/

int how_many_iter_real(double complex a, int maxiter)
{
  int i=0;
  double ar=creal(a);
  double ai=cimag(a);
  double br=0, bi=0;
  double tmp;
  double r;
  while (i<maxiter) {
    r=br*br+bi*bi;
    if (r>4)
      return i;
    tmp=br;
    br=br*br-bi*bi+ar;
    bi=ai+2*tmp*bi;
    i++;
  }
  return maxiter;
}

/*----------------------------------------------------------------------*/
void my_pgctab(float r,float R, float g, float G, float b, float B, int nc)
{
  float *rr=(float *)malloc(sizeof(float)*nc);
  float *gg=(float *)malloc(sizeof(float)*nc);
  float *bb=(float *)malloc(sizeof(float)*nc);
  float *ll=(float *)malloc(sizeof(float)*nc);
  for (int i=0;i<nc;i++) {
    rr[i]=r+(R-r)*(float)i/(float)(nc-1);
    gg[i]=g+(G-g)*(float)i/(float)(nc-1);
    bb[i]=b+(B-b)*(float)i/(float)(nc-1);
    ll[i]=(float)i/(float)(nc-1);
  }

  cpgctab(ll,rr,gg,bb,nc,1.0,0.5);

  free(rr);
  free(gg);
  free(bb);
  free(ll);
    
}

/*----------------------------------------------------------------------*/

float **make_mandel_map_omp(double xmin, double xmax, double ymin, double ymax, int npix, int maxiter)
{
  float **mymap=fmatrix(npix,npix);
  #pragma omp parallel shared(mymap,npix,xmin,xmax,ymin,ymax,maxiter) default(none) 
  {
    #pragma omp for 
     for (int i=0;i<npix;i++)  
        for (int j=0;j<npix;j++) {
           double x=((double)i)/((double)npix)*(xmax-xmin)+xmin;
           double y=((double)j)/((double)npix)*(ymax-ymin)+ymin;
           double complex a=x+I*y;
           mymap[i][j]=how_many_iter_real(a,maxiter);
        }
  }
  return mymap; 
}

/*----------------------------------------------------------------------*/

float **make_mandel_map(double xmin, double xmax, double ymin, double ymax, int npix, int maxiter)
{
  return make_mandel_map_omp(xmin,xmax,ymin,ymax,npix,maxiter);
}

/*======================================================================*/


int main(int argc, char *argv[])
{
  pca_time tt;
  int npix=800;
  int maxiter=1000;

  double xmin=-2;
  double xmax=2;
  double ymin=-2;
  double ymax=2;


  char pressed=' ';
  while (pressed!='X') {
    tick(&tt);
    float **map=make_mandel_map(xmin, xmax,ymin,ymax,npix,maxiter);
    tock(&tt);
    pressed=display_map_new_limits(map,npix,AUTOSCALE,AUTOSCALE,0.3,&xmin,&xmax,&ymin,&ymax);
    free(map[0]);
    free(map);
  }
}

